#!/usr/bin/env python3
# #
# @copyright Copyright (C) 2023 SpyroSoft Solutions S.A. All rights reserved.
# #

"""Generate toml for scargo project"""
import os

from jinja2 import Environment, FileSystemLoader

from scargo.scargo_src.global_values import SCARGO_PGK_PATH
from scargo.scargo_src.sc_logger import get_logger


class TomlTemplate:
    """
    This class is a container for env file which is used by docker compose
    Is providing basic environmental variable setup
    """

    def __init__(self, output_path, values):
        self.scargo_path = SCARGO_PGK_PATH
        self.output_path = output_path
        self.values = values
        self.jinja_env = Environment(
            loader=FileSystemLoader(os.path.join(self.scargo_path, "jinja/templates")),
            trim_blocks=True,
            lstrip_blocks=True,
        )

    def create_toml_file(self):
        """Function which creates a toml file using jinja"""
        with open(self.output_path, "w", encoding="utf-8") as out:
            out.write(
                self.jinja_env.get_template("scargo.toml.j2").render(data=self.values)
            )
        logger = get_logger()
        logger.info("Generated scargo.toml file")

    def create_toml(self):
        """Generate toml"""
        self.create_toml_file()
        return 0


def generate_toml(output_file_path, **values):
    """_summary_

    Args:
        output_file_path (String): path to the output .env file
        **values (Dict): dict contains all necessary values for toml

    Returns:
        bool: True is generation succeed
    """

    toml_template = TomlTemplate(output_file_path, values)
    status = toml_template.create_toml()

    return status



# #
# @copyright Copyright (C) 2023 SpyroSoft Solutions S.A. All rights reserved.
# #

"""Clean project from unnecessary files"""
import shutil
from pathlib import Path
from typing import Optional

from scargo.scargo_src.sc_logger import get_logger
from scargo.scargo_src.utils import get_project_root


def _case_insensitive_find_dir(source_dir: Path, dirname: str) -> Optional[Path]:
    if source_dir and source_dir.exists():
        for child in source_dir.iterdir():
            if child.name.lower() == dirname.lower():
                return child.absolute()
    return None


def scargo_clean() -> None:
    """Clean project dir from unnecessary files"""
    logger = get_logger()

    project_path = get_project_root()
    test_dir = _case_insensitive_find_dir(project_path, "test")
    source_directories = [project_path, test_dir]

    for source_dir in source_directories:
        build_dir = _case_insensitive_find_dir(source_dir, "build")
        if build_dir and build_dir.exists():
            shutil.rmtree(build_dir)
            logger.info("Removed %s", build_dir)



# #
# @copyright Copyright (C) 2023 SpyroSoft Solutions S.A. All rights reserved.
# #


def is_multiline(line):
    line = line.rstrip()
    if line.endswith("="):
        return True
    if line.count("(") != line.count(")"):
        return True
    return False


def missing_data_diff(old_data, new_data):
    """
    Prints missing mocks implementations and return True if there is any.
    """
    missing = False
    print("\033[31m")  # red color
    missing_mocks = list(
        set(new_data["missing_mocks"]) - set(old_data["missing_mocks"])
    )
    if missing_mocks:
        missing = True
        print("\nMissing mock file implementation:")
        for i in missing_mocks:
            print(f"    \u00B7 {i}")

    missing_classes = dict_diff(
        old_data["missing_classes"], new_data["missing_classes"]
    )
    if missing_classes:
        missing = True
        print("\nMissing mock class implementation:")
        for key, value in missing_classes.items():
            print(f"    \u00B7 {key}: {value}")

    missing_methods = dict_diff(
        old_data["missing_methods"], new_data["missing_methods"]
    )
    if missing_methods:
        missing = True
        print("\nMissing mock method implementation:")
        for key, value in missing_methods.items():
            print(f"    \u00B7 {key}: {value}")
    print("\033[0m")  # reset color

    return missing


def dict_diff(dict1, dict2):
    """
    Returns the difference between two dictionaries consisting of the key:list pair.
    """
    diffs = {}
    for key, value in dict2.items():
        if key not in dict1.keys():
            diffs[key] = value
            continue
        value_diff = set(value) - set(dict1[key])
        if value_diff:
            diffs[key] = value_diff
    return diffs



# #
# @copyright Copyright (C) 2023 SpyroSoft Solutions S.A. All rights reserved.
# #
"""Format project code using formatter"""
import os

from scargo.scargo_src.sc_src import (
    check_clang_format,
    check_copyright,
    check_pragma,
    prepare_config,
)
from scargo.scargo_src.utils import get_project_root


def scargo_fix(pragma: bool, copy_right: bool, clang_format: bool) -> None:
    """
    Fix format

    :param bool pragma: fix pragma format
    :param bool copy_right: fix copyrights
    :param bool clang_format: fix clang format
    :return: None
    """
    config = prepare_config()

    run_all = not any(
        [
            pragma,
            copy_right,
            clang_format,
        ]
    )

    # Todo, remove chdir and change cwd for checks
    os.chdir(get_project_root())

    if pragma or run_all:
        check_pragma(config, True)

    if copy_right or run_all:
        check_copyright(config, True)

    if clang_format or run_all:
        check_clang_format(config, True, False)



# #
# @copyright Copyright (C) 2023 SpyroSoft Solutions S.A. All rights reserved.
# #

from pathlib import Path

from jinja2 import Environment, FileSystemLoader

from scargo.scargo_src.sc_logger import get_logger
from scargo.scargo_src.sc_src import get_scargo_config_or_exit
from scargo.scargo_src.utils import get_config_file_path, get_project_root


class BaseGen:
    def __init__(self, templates_path: Path) -> None:
        self.jinja_env = Environment(
            loader=FileSystemLoader(templates_path),
            trim_blocks=True,
            lstrip_blocks=True,
        )

    def create_file_from_template(
        self,
        template: str,
        output_path: Path,
        overwrite: bool = True,
        **template_kwargs,
    ) -> None:
        """Function creates file using jinja template on output path, creates dirs if necessary"""
        if (
            self._is_file_excluded(output_path)
            or output_path.exists()
            and not overwrite
        ):
            return

        output_path.parent.mkdir(parents=True, exist_ok=True)
        content = self.jinja_env.get_template(template).render(template_kwargs)
        output_path.write_text(content, encoding="utf-8")

        logger = get_logger()
        logger.info("Generated %s", output_path)

    @staticmethod
    def _is_file_excluded(output_path: Path):
        # BaseGen is used in scargo new as well as scargo update, so sometimes
        # scargo.lock does not exist yet. Take values from toml instead.
        config_path = get_config_file_path("scargo.lock") or get_config_file_path(
            "scargo.toml"
        )
        toml_data = get_scargo_config_or_exit(config_path)
        exclude_list = toml_data.scargo.update_exclude
        project_path = get_project_root()
        return str(output_path.absolute().relative_to(project_path)) in exclude_list



# #
# @copyright Copyright (C) 2023 SpyroSoft Solutions S.A. All rights reserved.
# #

import subprocess
from pathlib import Path

from scargo.scargo_src.sc_config import Config
from scargo.scargo_src.sc_logger import get_logger
from scargo.scargo_src.sc_src import prepare_config
from scargo.scargo_src.utils import get_project_root


def scargo_flash(app: bool, fs: bool, flash_profile: str) -> None:
    config = prepare_config()
    target = config.project.target

    if target.family == "esp32":
        flash_esp32(config, app=app, fs=fs, flash_profile=flash_profile)
    elif target.family == "stm32":
        flash_stm32(config, flash_profile)
    else:
        logger = get_logger()
        logger.error("Flash command not supported for this target yet.")


def flash_esp32(
    config: Config, app: bool, fs: bool, flash_profile: str = "Debug"
) -> None:
    project_path = get_project_root()
    out_dir = project_path / "build" / flash_profile
    target = config.project.target
    command = ""
    try:
        if app:
            app_name = config.project.name
            app_path = out_dir / f"{app_name}.bin"
            command = (
                f"parttool.py write_partition --partition-name=ota_0 --input {app_path}"
            )
            subprocess.check_call(command, shell=True, cwd=project_path)
        elif fs:
            fs_path = Path("build") / "spiffs.bin"
            command = (
                f"parttool.py write_partition --partition-name=spiffs --input {fs_path}"
            )
            subprocess.check_call(command, shell=True, cwd=project_path)
        else:
            command = f"esptool.py --chip {target.id} write_flash @flash_args"
            subprocess.check_call(command, shell=True, cwd=out_dir)
    except subprocess.CalledProcessError:
        logger = get_logger()
        logger.error("%s fail", command)


def flash_stm32(config: Config, flash_profile: str = "Debug") -> None:
    logger = get_logger()

    project_path = get_project_root()
    bin_name = config.project.bin_name.lower()
    bin_path = project_path / "build" / flash_profile / "bin" / f"{bin_name}.bin"

    flash_start = hex(config.stm32.flash_start)

    if not bin_path.exists():
        logger.error("%s does not exist", bin_path)
        logger.info("Did you run scargo build --profile %s", flash_profile)
    elif not flash_start:
        logger.error("Flash start address not found in lock file")
        logger.info("Define flash-start in scargo.toml under stm32 section")
    else:
        subprocess.check_call(["st-flash", "erase"])
        subprocess.check_call(
            ["st-flash", "--reset", "write", str(bin_path), flash_start]
        )



#!/usr/bin/env python3
# #
# @copyright Copyright (C) 2023 SpyroSoft Solutions S.A. All rights reserved.
# #

import json
import os
from argparse import ArgumentParser
from pathlib import Path

import clang.cindex as cindex
from jinja2 import Environment, FileSystemLoader

from scargo.jinja.mock_utils.cmake_utlis import add_subdirs_to_cmake, create_cmake_lists
from scargo.jinja.mock_utils.data_classes import (
    ArgumentDescriptor,
    HeaderDescriptor,
    MockClassDescriptor,
    MockFunctionDescriptor,
    MockNamespaceDescriptor,
)
from scargo.jinja.mock_utils.error_classes import MockNotImplemented
from scargo.jinja.mock_utils.scraping_utils import missing_data_diff

# define paths
absolute_path = Path(__file__).parent.absolute()
SRC_DIR = "src"
MOCKS_DIR = "tests/mocks"

jinja_env = Environment(
    loader=FileSystemLoader(os.path.join(absolute_path, "mock_templates")),
    trim_blocks=True,
    lstrip_blocks=True,
)

# json where the mock paths are held, which lack of implementation has been accepted
missing_mocks_json = os.path.join(absolute_path, "mock_utils", "missing_mocks.json")


def option_parser_init():
    """This function parses the arguments which are given and
    returns known arguments for path and exclude and unknown arguments"""

    parser = ArgumentParser(epilog="The script creates mock files for cpp code.")

    parser.add_argument(
        "-p",
        "--path_to_src",
        type=str,
        default=SRC_DIR,
        help="Path to source directory or single header file",
    )

    parser.add_argument(
        "-e",
        "--exclude",
        default=[],
        action="append",
        help="Exclude directories",
    )

    parser.add_argument(
        "-f",
        "--force",
        action="store_true",
        help="Force mock folder override",
    )

    parser.add_argument(
        "-c",
        "--check",
        action="store_true",
        help="Check if the mocks exist",
    )

    args = parser.parse_args()
    return args


def header_files_in_path(path_to_src, exclude, force, check_only, src_dir=SRC_DIR):
    """
    Returns list of the .h files in the path.
    If not check_only it also adds paths to the CMakelist.
    """
    header_files = []

    if os.path.isfile(path_to_src):
        root, file = os.path.split(path_to_src)
        mock_root = root.replace(src_dir, MOCKS_DIR, 1)
        if file.endswith(".h") or file.endswith(".hpp"):
            header_files.append((root, mock_root, file))
        else:
            print("Not a header file. Please chose .h or .hpp file.")

    if os.path.isdir(path_to_src):
        for root, _, files in os.walk(path_to_src):
            mock_root = root.replace(src_dir, MOCKS_DIR, 1)

            # Skip directories in the excluded directory list
            if exclude and root in exclude:
                print("Excluded:", root)
                continue

            for file in files:
                # We only consider files ending in .h
                if file.endswith(".h") or file.endswith(".hpp"):
                    header_files.append((root, mock_root, file))

            if check_only:
                continue

            # if tests/mocks in path doesn't exist then create it
            if not os.path.exists(mock_root):
                os.makedirs(mock_root)

            # generate CMakeLists if it does not exist or if forced
            if not os.path.isfile(mock_root + "/CMakeLists.txt") or force:
                # create CMakeLists file in root
                create_cmake_lists(mock_root, jinja_env)

            # add subdirs to CMakeLists
            add_subdirs_to_cmake(mock_root)

    return header_files


def check_mocks(path_to_src, exclude, force, check_only):
    """
    Check if there is a mock implementations for cpp files in the path.
    """
    src_dir = "main" if "main" in path_to_src else SRC_DIR
    header_files = header_files_in_path(
        path_to_src, exclude, force, check_only, src_dir
    )
    missing_mocks = []
    missing_classes = {}
    missing_methods = {}

    for root, mock_root, file in header_files:
        src_header = os.path.join(root, file)
        dst_header = os.path.join(mock_root, file)

        # Skip files in the excluded list
        if exclude and src_header in exclude:
            print("Excluded:", src_header)
            continue

        # Scraping data from exist source and mock files to compare
        if os.path.isfile(dst_header):
            # Read classes and functions from the source header
            header_data = parse_file(src_header)
            # Check classes and methods in mock file
            mock_file = open(dst_header, "r", encoding="utf-8").read()
            for mock_class in header_data.classes:
                if mock_class.name and mock_class.name not in mock_file:
                    missing_classes.setdefault(src_header, []).append(mock_class.name)
                    continue
                for method in mock_class.methods:
                    if method.name not in mock_file:
                        missing_methods.setdefault(src_header, []).append(method.name)
            for mock_class in header_data.one_line_classes:
                mock_class = mock_class.strip().replace(";", "")
                if mock_class not in mock_file:
                    missing_classes.setdefault(src_header, []).append(mock_class)
        else:
            missing_mocks.append(src_header)

    data = {
        "missing_mocks": missing_mocks,
        "missing_classes": missing_classes,
        "missing_methods": missing_methods,
    }

    with open(missing_mocks_json, "r", encoding="utf-8") as json_file:
        old_data = json.load(json_file)

    if missing_data_diff(old_data, data):
        raise MockNotImplemented()


def generate_mocks(path_to_src: Path, exclude=(), force=False, check_only=False):
    """
    Generates mock header and implementations for specified source headers.
    Creates directories and CMake lists where required.
    :param path_to_src Path to source directory or header
    :param exclude List of files or directories to exclude from mock generator
    :param force Force overwrite
    :param check_only
    """
    src_dir = path_to_src.parent.name
    header_files = header_files_in_path(
        path_to_src, exclude, force, check_only, src_dir
    )

    for root, mock_root, file in header_files:

        src_header = os.path.join(root, file)
        dst_header = os.path.join(mock_root, file)
        dst_mock = os.path.join(mock_root, f"mock_{file}")

        # Skip files in the excluded list
        if exclude and src_header in exclude:
            print("Excluded:", src_header)
            continue

        # Skip generation if destination header exists
        if os.path.isfile(dst_header):
            print("Skipping:", src_header)
            continue

        # Read classes and functions from the source header
        header_data = parse_file(src_header)

        # generate public interface header
        with open(dst_header, "w", encoding="utf-8") as out:
            out.write(
                jinja_env.get_template("class_interface.h.j2").render(
                    header=header_data
                )
            )

        # generate mock header
        with open(dst_mock, "w", encoding="utf-8") as out:
            out.write(
                jinja_env.get_template("class_mock.h.j2").render(header=header_data)
            )

        # generate mock implementation
        with open(dst_mock.replace(".h", ".cpp"), "w", encoding="utf-8") as out:
            out.write(
                jinja_env.get_template("class_mock.cpp.j2").render(header=header_data)
            )

        print("Generated:", src_header)


def find_namespaces_in_file(cursor, filename):
    for i in cursor.walk_preorder():
        if i.kind != cindex.CursorKind.NAMESPACE:
            continue
        if i.location.file.name != filename:
            continue
        yield i.spelling


class ParamsExtractor(object):
    @staticmethod
    def extract_params_of_children_in_file(cursor, type_list, filename=None):
        return [
            ParamsExtractor.extract_params(x)
            for x in ParamsExtractor.find_children_of_type_in_file(
                cursor, type_list, filename
            )
        ]

    @staticmethod
    def find_children_of_type_in_file(cursor, type_list, filename=None):
        for i in cursor.walk_preorder():
            if i.kind not in type_list:
                continue
            if filename is not None and i.location.file.name != filename:
                continue
            yield i

    @staticmethod
    def extract_params(cursor):
        if cursor.kind == cindex.CursorKind.FUNCTION_DECL:
            return __class__._extract_function_params(cursor)
        elif cursor.kind == cindex.CursorKind.CXX_METHOD:
            return __class__._extract_method_params(cursor)
        elif cursor.kind == cindex.CursorKind.CONSTRUCTOR:
            return __class__._extract_method_params(cursor)
        elif cursor.kind == cindex.CursorKind.DESTRUCTOR:
            return __class__._extract_method_params(cursor)
        elif cursor.kind == cindex.CursorKind.CLASS_DECL:
            return __class__._extract_class_params(cursor)
        elif cursor.kind == cindex.CursorKind.NAMESPACE:
            return __class__._extract_namespace_params(cursor)

    @staticmethod
    def _extract_method_params(cursor):
        fun_ret_val = cursor.result_type.spelling
        fun_name = cursor.spelling
        fun_args = []
        specifiers = []

        if cursor.access_specifier.name != "PUBLIC":
            return
        if cursor.is_virtual_method():
            specifiers.append("override")
        if cursor.is_const_method():
            specifiers.append("const")
        for i in cursor.walk_preorder():
            if i.kind == cindex.CursorKind.PARM_DECL:
                fun_args.append(ArgumentDescriptor(i.spelling, i.type.spelling))
        return MockFunctionDescriptor(
            fun_name, fun_ret_val, specifiers, arguments=fun_args
        )

    @staticmethod
    def _extract_function_params(cursor):
        fun_ret_val = cursor.result_type.spelling
        fun_name = cursor.spelling
        fun_args = []
        for i in cursor.walk_preorder():
            if i.kind == cindex.CursorKind.PARM_DECL:
                fun_args.append(ArgumentDescriptor(i.spelling, i.type.spelling))
        return MockFunctionDescriptor(fun_name, fun_ret_val, "", arguments=fun_args)

    @staticmethod
    def _extract_namespace_params(cursor):
        return MockNamespaceDescriptor(cursor.spelling)

    @staticmethod
    def _extract_class_params(cursor):
        cls = MockClassDescriptor(cursor.spelling, f"Mock{cursor.spelling}")

        methods = ParamsExtractor.find_children_of_type_in_file(
            cursor, [cindex.CursorKind.CXX_METHOD]
        )

        cls.methods = [ParamsExtractor.extract_params(m) for m in methods]

        # delete Nones from lists
        cls.methods = [x for x in cls.methods if x is not None]
        return cls


def parse_file(file_path):
    """TODO documentation
    look here for available types which can be extracted
    https://clang.llvm.org/doxygen/group__CINDEX.html
    This function receives a path with file name as an argument and returns
    list of includes, i.e. every string that starts with "#include",
    dict of classes where keys are "public", "private" and "mock_class_name"
    and list of namespaces, i.e. strings that start with "namespace".
    """
    file_name = os.path.basename(file_path)

    hdr = HeaderDescriptor(file_name)

    idx = cindex.Index.create()
    translation_unit = idx.parse(file_path, ["-x", "c++"])

    hdr.namespaces = ParamsExtractor.extract_params_of_children_in_file(
        translation_unit.cursor, [cindex.CursorKind.NAMESPACE], file_path
    )

    # defns = params_extractor.find_children_of_type_in_file(
    #     translation_unit.cursor, [cindex.CursorKind.FUNCTION_DECL], file_path
    # )

    hdr.classes = ParamsExtractor.extract_params_of_children_in_file(
        translation_unit.cursor, [cindex.CursorKind.CLASS_DECL], file_path
    )
    return hdr


def main():
    args = option_parser_init()

    if args.check and args.force:
        raise ValueError("You shouldn't use the check and force option together.")
    if args.check:
        check_mocks(args.path_to_src, args.exclude, args.force, args.check)
        pass
    else:
        generate_mocks(args.path_to_src, args.exclude, args.force, args.check)


if __name__ == "__main__":
    main()



# #
# @copyright Copyright (C) 2023 SpyroSoft Solutions S.A. All rights reserved.
# #

import subprocess
import sys
from pathlib import Path
from time import sleep
from typing import Optional

from scargo.scargo_src.sc_config import Config
from scargo.scargo_src.sc_logger import get_logger
from scargo.scargo_src.sc_src import prepare_config
from scargo.scargo_src.utils import find_program_path, get_project_root


class _ScargoDebug:
    SUPPORTED_TARGETS = ["x86", "stm32"]

    def __init__(self, config: Config, bin_path: Optional[Path]):
        self._logger = get_logger()
        self._target = config.project.target

        if self._target.family not in self.SUPPORTED_TARGETS:
            self._logger.error("Debugging currently not supported for %s", self._target)
            self._logger.info(
                "Scargo currently supports debug for %s", self.SUPPORTED_TARGETS
            )
            sys.exit(1)

        self._bin_path = bin_path or self._get_bin_path(config.project.bin_name)
        if not self._bin_path.exists():
            self._logger.error("Binary %s does not exist", self._bin_path)
            self._logger.info("Did you run scargo build --profile Debug?")
            sys.exit(1)

        if self._target.family == "stm32":
            self._chip = config.stm32.chip
            if not self._chip:
                self._logger.error("Chip label not defined in toml.")
                self._logger.info(
                    "Define chip under stm32 section and run scargo update."
                )
                sys.exit(1)

    def run_debugger(self):
        if self._target.family == "x86":
            self._debug_x86()
        elif self._target.family == "stm32":
            self._debug_stm32()

    def _debug_x86(self):
        subprocess.run(f"gdb {self._bin_path}", shell=True)

    def _debug_stm32(self):
        openocd_path = find_program_path("openocd")
        if not openocd_path:
            self._logger.error("Could not find openocd.")
            sys.exit(1)

        chip_script = f"target/{self._chip[:7].lower()}x.cfg"
        openocd = subprocess.Popen(
            [
                openocd_path,
                "-f",
                "interface/stlink-v2-1.cfg",
                "-f",
                chip_script,
                "-f",
                ".devcontainer/stm32.cfg",
            ]
        )
        # Wait for openocd to start
        sleep(1)
        try:
            subprocess.run(
                [
                    "gdb-multiarch",
                    self._bin_path,
                    '--eval-command="target extended-remote localhost:3333"',
                ],
            )
        finally:
            openocd.terminate()

    def _get_bin_path(self, bin_name: str) -> Path:
        project_path = get_project_root()
        bin_path = Path(project_path, "build/Debug/bin", bin_name).absolute()
        if self._target.family == "stm32" and bin_path.suffix != "elf":
            bin_path = bin_path.with_suffix(".elf")
        return bin_path


def scargo_debug(bin_path: Optional[Path]):
    config = prepare_config()
    sc_debug = _ScargoDebug(config, bin_path)
    sc_debug.run_debugger()



# #
# @copyright Copyright (C) 2023 SpyroSoft Solutions S.A. All rights reserved.
# #

from pathlib import Path

from scargo.jinja.base_gen import BaseGen
from scargo.scargo_src.global_values import SCARGO_PGK_PATH
from scargo.scargo_src.sc_config import Config, Target
from scargo.scargo_src.utils import get_project_root


class _CppTemplateGen(BaseGen):
    """
    This class is a container cpp files creation with multilayer approach
    """

    def __init__(self):
        self.template_dir = Path(SCARGO_PGK_PATH, "jinja", "cpp")
        BaseGen.__init__(self, self.template_dir)

    def _set_src_dir(self, target: Target):
        self._src_dir = get_project_root() / target.source_dir

    def _generate_bin(self, target: str, bin_name: str):
        """Function which creates main.cpp file using jinja"""
        self.create_file_from_template(
            "main.cpp.j2",
            Path(self._src_dir, f"{bin_name.lower()}.cpp"),
            target=target,
            bin_name=bin_name,
        )

    def _generate_lib(self, lib_name: str):
        """Function which creates a lib files using jinja"""
        self.create_file_from_template(
            "lib.cpp.j2",
            Path(self._src_dir, f"{lib_name.lower()}.cpp"),
            lib_name=lib_name,
        )
        self.create_file_from_template(
            "lib.h.j2", Path(self._src_dir, f"{lib_name.lower()}.h"), lib_name=lib_name
        )

    def _generate_cmake(self, config: Config):
        target = config.project.target

        self.create_file_from_template(
            f"cmake-src-{target.family}.j2",
            Path(self._src_dir, "CMakeLists.txt"),
            config=config,
        )

    def generate_cpp(self, config: Config):
        """Generate dirs and files"""
        target = config.project.target
        self._set_src_dir(target)

        lib_name = config.project.lib_name
        if lib_name:
            self._generate_lib(lib_name)

        bin_name = config.project.bin_name
        if bin_name:
            self._generate_bin(target, bin_name)

        self._generate_cmake(config)


def generate_cpp(config: Config):
    cpp_template = _CppTemplateGen()
    cpp_template.generate_cpp(config)



# #
# @copyright Copyright (C) 2023 SpyroSoft Solutions S.A. All rights reserved.
# #

"""Update project"""
import shutil
import subprocess
from pathlib import Path

from scargo import __version__ as ver
from scargo.jinja.cicd_gen import generate_cicd
from scargo.jinja.cmake_gen import generate_cmake
from scargo.jinja.conan_gen import generate_conanfile
from scargo.jinja.docker_gen import generate_docker_compose
from scargo.jinja.env_gen import generate_env
from scargo.jinja.readme_gen import generate_readme
from scargo.jinja.tests_gen import generate_tests
from scargo.scargo_src.global_values import (
    SCARGO_DOCKER_ENV,
    SCARGO_LOCK_FILE,
    SCARGO_PGK_PATH,
)
from scargo.scargo_src.sc_docker import scargo_docker
from scargo.scargo_src.sc_logger import get_logger
from scargo.scargo_src.sc_src import check_scargo_version, get_scargo_config_or_exit
from scargo.scargo_src.utils import get_project_root


def copy_file_if_not_exists() -> None:
    """
    Copy file from scargo pkg

    :return: None
    """
    files_to_copy = Path(SCARGO_PGK_PATH, "templates").glob("*")
    project_abs_path = get_project_root()
    for file in files_to_copy:
        if not Path(project_abs_path, file.name).exists():
            shutil.copy2(file, project_abs_path)


def scargo_update(config_file_path: Path) -> None:
    """
    Update project

    :param config_file_path: path to .toml configuration file (e.g. scargo.toml)
    :return: None
    """
    logger = get_logger()
    project_path = get_project_root()
    docker_path = Path(project_path, ".devcontainer")
    config = get_scargo_config_or_exit(config_file_path)
    if not config.project:
        logger.error("File `%s`: Section `project` not found.", config_file_path)
        return

    if not config.project.name:
        logger.error(
            "File `{config_file_path}`: `name` not found under `project` section."
        )
        return

    # Copy templates project files to repo directory
    copy_file_if_not_exists()

    # Copy config file and create lock file.
    shutil.copyfile(config_file_path, project_path / SCARGO_LOCK_FILE)
    ###########################################################################
    config = get_scargo_config_or_exit()
    check_scargo_version(config)
    project_config = config.project
    target = project_config.target

    # Copy docker env files to repo directory
    generate_docker_compose(docker_path, project_config, ver)
    generate_env(docker_path)

    generate_cmake(config)
    generate_conanfile(config)

    if target.family == "esp32":
        Path(target.source_dir, "fs").mkdir(parents=True, exist_ok=True)
        with open(Path(project_path, "version.txt"), "w", encoding="utf-8") as out:
            out.write(project_config.version)
        with open(Path(project_path, "partitions.csv"), "w", encoding="utf-8") as out:
            out.write(
                "# ESP-IDF Partition Table\n# Name,   Type, SubType, Offset,  Size, Flags\n"
            )
            partitions = config.esp32.partitions
            for line in partitions:
                out.write(line + "\n")
            out.write("\n")

    generate_cicd(project_config=project_config)
    generate_tests(target, config.tests)
    generate_readme(project_config)

    # do not rebuild dockers in the docker
    if target.family == "stm32" and not Path("third-party/stm32-cmake").is_dir():
        subprocess.run("conan source .", shell=True, cwd=project_path)

    if project_config.build_env == SCARGO_DOCKER_ENV:
        scargo_docker(build_docker=True)



# #
# @copyright Copyright (C) 2023 SpyroSoft Solutions S.A. All rights reserved.
# #
"""feature function for scargo"""
import glob
import os
import subprocess
import sys
from pathlib import Path
from typing import Optional, Tuple

import docker as dock
import tomlkit

from scargo import __version__ as ver
from scargo.scargo_src.global_values import SCARGO_DOCKER_ENV, SCARGO_LOCK_FILE
from scargo.scargo_src.sc_config import Config, ProjectConfig, Target, parse_config
from scargo.scargo_src.sc_logger import get_logger
from scargo.scargo_src.utils import get_config_file_path, get_project_root

###############################################################################


def run_scargo_again_in_docker(project_config: ProjectConfig) -> None:
    """
    Run command in docker

    :param dict project_config: project configuration
    :return: None
    """
    project_path = get_project_root()
    relative_path = Path.cwd().absolute().relative_to(project_path)

    cmd_args = " ".join(sys.argv[1:])

    entrypoint = ""
    if project_config.target.family == "esp32":
        entrypoint = "/opt/esp/entrypoint.sh"

    cmd = f'/bin/bash -c "scargo version || true; cd {relative_path} && scargo {cmd_args}"'

    docker_tag = project_config.docker_image_tag
    logger = get_logger()

    if project_path:
        logger.info("Running scargo %s command in docker.", cmd_args)

        client = dock.from_env()
        container = client.containers.run(
            f"{docker_tag}",
            cmd,
            volumes=[str(project_path) + ":/workspace/", "/dev/:/dev/"],
            entrypoint=entrypoint,
            privileged=True,
            detach=True,
            remove=True,
        )
        # make life output
        output = container.attach(stdout=True, stream=True, logs=True)
        for line in output:
            print(line.decode(), end="")
        # exit as command already processed in docker
        sys.exit(0)


###############################################################################


def check_scargo_version(config: Config) -> None:
    """
    Check scargo version

    :param Config config: project configuration
    :return: None
    """
    version_lock = config.scargo.version
    if not version_lock:
        add_version_to_scargo_lock()
    elif ver != version_lock:
        logger = get_logger()
        logger.warning("Warning: scargo package is different then in lock file")
        logger.info("Run scargo update")


###############################################################################


def get_scargo_config_or_exit(
    config_file_path: Optional[Path] = None,
) -> Config:
    """
    :param config_file_path
    :return: project configuration as dict
    """
    if config_file_path is None:
        config_file_path = get_config_file_path(SCARGO_LOCK_FILE)
    if config_file_path is None or not config_file_path.exists():
        logger = get_logger()
        logger.error("File `%s` does not exist.", SCARGO_LOCK_FILE)
        logger.info("Did you run `scargo update`?")
        sys.exit(1)

    return parse_config(config_file_path)


###############################################################################


def add_version_to_scargo_lock() -> None:
    """
    :return: project configuration as dict
    """
    scargo_lock = get_config_file_path(SCARGO_LOCK_FILE)
    if not scargo_lock:
        logger = get_logger()
        logger.error("ERROR: File `%s` does not exist.", SCARGO_LOCK_FILE)
        logger.info("Did you run `scargo update`?")
        sys.exit(1)

    scargo_lock_file = open(scargo_lock)
    config = tomlkit.load(scargo_lock_file)

    config.setdefault("scargo", tomlkit.table())["version"] = ver
    scargo_lock_file = open(scargo_lock, "w")
    tomlkit.dump(config, scargo_lock_file)


###############################################################################


def get_docker_files_from_scargo_pkg(directory: Path, target: Target) -> list:
    """
    Copy docker file from scargo pkg

    :param Path directory: output dir
    :param target: type of architecture
    :return: list of files to copy
    """
    available_dir_list = ["base", "user", "cpp", target.id]
    result_list_of_files = []
    files = Path(directory).glob("*")
    for file in files:
        if Path(file).is_dir():
            if Path(file).name in available_dir_list:
                result_list_of_files.append(str(file))
    return result_list_of_files


###############################################################################


def check_pragma(config: Config, fix_errors: bool) -> None:
    """
    Private function used in commands `scargo check` and `scargo fix`.

    :param Config config: project configuration
    :param bool fix_errors: if fix errors during code checking
    :return:
    """

    logger = get_logger()
    logger.info("Starting pragma check...")

    file_extensions = (".h", ".hpp")
    error_counter = 0
    source_dir = config.project.target.source_dir

    exclude_list = []

    # Collect global excludes.
    for pattern in config.check.exclude:
        exclude_list.extend(glob.glob(pattern))

    # Collect local excludes.
    for pattern in config.check.pragma.exclude:
        exclude_list.extend(glob.glob(pattern))

    for root, _, filenames in os.walk(source_dir):
        for fname in filenames:
            fname = os.path.join(root, fname)

            exist_flag = False
            for exclude in exclude_list:
                if exclude in fname:
                    logger.info("Skipping %s", fname)
                    exist_flag = True
                    break
            if exist_flag:
                continue

            if fname.lower().endswith(file_extensions):
                found_pragma_once = False

                with open(fname, encoding="utf-8") as file:
                    print(fname)
                    for line in file.readlines():
                        if "#pragma once" in line:
                            found_pragma_once = True
                            break

                if not found_pragma_once:
                    error_counter += 1
                    logger.warning("Missing pragma in %s", fname)

                    if fix_errors:
                        logger.info("Fixing...")

                        with open(fname, encoding="utf-8") as file:
                            old = file.read()

                        with open(fname, "w", encoding="utf-8") as file:
                            file.write("#pragma once\n")
                            file.write("\n")
                            file.write(old)

    if fix_errors:
        logger.info("Finished pragma check. Fixed problems in %s files.", error_counter)
    else:
        logger.info("Finished pragma check. Found problems in %s files.", error_counter)
        if error_counter > 0:
            logger.error("pragma check fail!")
            sys.exit(1)


###############################################################################


def check_copyright(config: Config, fix_errors: bool) -> None:
    """
    Private function used in commands `scargo check` and `scargo fix`.

    :param Config config: project configuration
    :param bool fix_errors: if fix errors during code checking
    :return:
    """
    logger = get_logger()
    logger.info("Starting copyright check...")

    file_extensions = (".h", ".hpp", ".c", ".cpp")
    error_counter = 0
    source_dir = config.project.target.source_dir

    exclude_list = []
    copyright_desc = config.check.copyright.description
    if not copyright_desc:
        logger.warning("No copyrights in defined in toml")
        return

    # Collect global excludes.
    for pattern in config.check.exclude:
        exclude_list.extend(glob.glob(pattern))

    # Collect local excludes.
    for pattern in config.check.copyright.exclude:
        exclude_list.extend(glob.glob(pattern))

    for root, _, filenames in os.walk(source_dir):
        for fname in filenames:
            fname = os.path.join(root, fname)

            exist_flag = False
            for exclude in exclude_list:
                if exclude in fname:
                    logger.info("Skipping %s", fname)
                    exist_flag = True
                    break
            if exist_flag:
                continue

            if fname.lower().endswith(file_extensions):
                found_copyright = False
                any_copyright = False

                with open(fname, encoding="utf-8") as file:
                    for line in file.readlines():
                        if copyright_desc in line:
                            found_copyright = True
                            break
                        if "copyright" in line.lower():
                            any_copyright = True
                            break

                if not found_copyright:
                    error_counter += 1
                    if any_copyright:
                        logger.warning(
                            "Incorrect and not excluded copyright in %s", fname
                        )
                    else:
                        logger.info("Missing copyright in %s.", fname)

                    if fix_errors and not any_copyright:
                        logger.info("Fixing...")
                        with open(fname, encoding="utf-8") as file:
                            old = file.read()

                        with open(fname, "w", encoding="utf-8") as file:
                            file.write("//\n")
                            file.write(f"// {copyright_desc}\n")
                            file.write("//\n")
                            file.write("\n")
                            file.write(old)

    if fix_errors:
        logger.info(
            "Finished copyright check. Fixed problems in %s files.", error_counter
        )
    else:
        logger.info(
            "Finished copyright check. Found problems in %s files.", error_counter
        )
        if error_counter > 0:
            logger.error("copyright check fail!")
            sys.exit(1)


###############################################################################


def check_todo(config: Config) -> None:
    """
    Private function used in command `scargo check`.

    :param Config config: project configuration
    :return: None
    """

    logger = get_logger()
    logger.info("Starting todo check...")

    keywords = ("tbd", "todo", "TODO", "fixme")
    file_extensions = (".h", ".hpp", ".c", ".cpp")
    error_counter = 0
    source_dir = config.project.target.source_dir

    exclude_list = []

    # Collect global excludes.
    for pattern in config.check.exclude:
        exclude_list.extend(glob.glob(pattern))

    # Collect local excludes.
    for pattern in config.check.todo.exclude:
        exclude_list.extend(glob.glob(pattern))

    for root, _, filenames in os.walk(source_dir):
        for fname in filenames:
            fname = os.path.join(root, fname)

            exist_flag = False
            for exclude in exclude_list:
                if exclude in fname:
                    logger.info("Skipping %s", fname)
                    exist_flag = True
                    break
            if exist_flag:
                continue

            if fname.lower().endswith(file_extensions):
                with open(fname, encoding="utf-8") as file:
                    line_number = 0
                    for line in file.readlines():
                        line_number += 1
                        for keyword in keywords:
                            if keyword in line:
                                error_counter += 1
                                logger.warning(
                                    f"Found {keyword} in {fname} at line {line_number}"
                                )

    logger.info("Finished todo check. Found %s problems.", error_counter)
    if error_counter > 0:
        logger.error("todo check fail!")
        sys.exit(1)


###############################################################################


def check_clang_format(config: Config, fix_errors: bool, verbose: bool) -> None:
    """
    Private function used in commands `scargo check` and `scargo fix`.

    :param Config config: project configuration
    :param bool fix_errors: if fix errors during check
    :param bool verbose: if verbose
    :return: None
    """
    logger = get_logger()
    logger.info("Starting clang-format check...")

    file_extensions = (".h", ".hpp", ".c", ".cpp")
    error_counter = 0
    source_dir = config.project.target.source_dir

    exclude_list = []

    # Collect global excludes.
    for pattern in config.check.exclude:
        exclude_list.extend(glob.glob(pattern))

    # Collect local excludes.
    for pattern in config.check.clang_format.exclude:
        exclude_list.extend(glob.glob(pattern))

    for root, _, filenames in os.walk(source_dir):
        for fname in filenames:
            fname = os.path.join(root, fname)

            exist_flag = False
            for exclude in exclude_list:
                if exclude in fname:
                    logger.info("Skipping %s", fname)
                    exist_flag = True
                    break
            if exist_flag:
                continue

            if fname.lower().endswith(file_extensions):
                cmd = "clang-format -style=file --dry-run " + fname
                out = subprocess.getoutput(cmd)

                if out != "":
                    error_counter += 1

                    if verbose:
                        logger.info(out)
                    else:
                        logger.warning("clang-format found error in file %s", fname)

                    if fix_errors:
                        logger.info("Fixing...")

                        cmd = "clang-format -style=file -i " + fname
                        subprocess.check_call(cmd, shell=True)

    if fix_errors:
        logger.info(
            "Finished clang-format check. Fixed problems in %s files.", error_counter
        )
    else:
        logger.info(
            "Finished clang-format check. Found problems in %s files.", error_counter
        )
        if error_counter > 0:
            logger.error("clang-format check fail!")
            sys.exit(1)


###############################################################################


def check_clang_tidy(config: Config, verbose: bool) -> None:
    """
    Private function used in command `scargo check`.

    :param Config config: project configuration
    :param bool verbose: if verbose
    :return: None
    """
    logger = get_logger()
    logger.info("Starting clang-tidy check...")

    file_extensions = (".h", ".hpp", ".c", ".cpp")
    error_counter = 0
    source_dir = config.project.target.source_dir

    exclude_list = []

    # Collect global excludes.
    for pattern in config.check.exclude:
        exclude_list.extend(glob.glob(pattern))

    # Collect local excludes.
    for pattern in config.check.clang_tidy.exclude:
        exclude_list.extend(glob.glob(pattern))

    for root, _, filenames in os.walk(source_dir):
        for fname in filenames:
            fname = os.path.join(root, fname)

            exist_flag = False
            for exclude in exclude_list:
                if exclude in fname:
                    logger.info("Skipping %s", fname)
                    exist_flag = True
                    break
            if exist_flag:
                continue

            if fname.lower().endswith(file_extensions):
                cmd = "clang-tidy " + fname + " --assume-filename=.hxx --"
                out = subprocess.getoutput(cmd)

                if "error:" in out:
                    error_counter += 1

                    if verbose:
                        logger.info(out)
                    else:
                        logger.warning("clang-tidy found error in file %s", fname)

    logger.info("Finished clang-tidy check. Found problems in %s files.", error_counter)
    if error_counter > 0:
        logger.error("clang-tidy check fail!")
        sys.exit(1)


###############################################################################


def check_cyclomatic(config: Config) -> None:
    """
    Private function used in command `scargo check`.

    :param Config config: project configuration
    :return: None
    :raises CalledProcessError: if cyclomatic check fail
    """

    logger = get_logger()
    logger.info("Starting cyclomatic check...")

    source_dir = config.project.target.source_dir

    # Collect global excludes.
    exclude_list = config.check.exclude

    # Collect local excludes.
    exclude_list.extend(config.check.cyclomatic.exclude)

    cmd = "lizard " + source_dir + " -C 25 -w"

    for exclude_pattern in exclude_list:
        cmd = cmd + " --exclude " + exclude_pattern
    try:
        subprocess.check_call(cmd, shell=True)
    except subprocess.CalledProcessError:
        logger.error("ERROR: Check cyclomatic fail")
    logger.info("Finished cyclomatic check.")


###############################################################################


def check_cppcheck() -> None:
    """
    Private function used in command `scargo check`.

    :return: None
    """
    logger = get_logger()
    logger.info("Starting cppcheck check...")

    cmd = "cppcheck --enable=all --suppress=missingIncludeSystem src/ main/"
    try:
        subprocess.check_call(cmd, shell=True)
    except subprocess.CalledProcessError:
        logger.error("cppcheck fail")
    logger.info("Finished cppcheck check.")


###############################################################################


def prepare_config() -> Config:
    """
    Prepare configuration file

    :return: project configuration
    """
    config = get_scargo_config_or_exit()
    check_scargo_version(config)
    ###########################################################################
    project_config = config.project
    build_env = project_config.build_env
    if build_env == SCARGO_DOCKER_ENV and not os.path.exists("/.dockerenv"):
        run_scargo_again_in_docker(project_config)
    ###########################################################################
    return config


def get_cc_config(target: Target) -> Tuple[str, str, str, str]:
    """
    Get c configuration base on architecture

    :param target: project architecture
    :return: tuple of string
    :raises Exception: if architecture not allowed
    """
    cflags = "-Wall -Wextra"
    cxxflags = "-Wall -Wextra"
    cc = target.cc
    cxx = target.cxx
    return cc, cflags, cxx, cxxflags


def get_build_env(create_docker: bool) -> str:
    """
    Get build env
    :param bool create_docker: if create docker
    :return: build env
    """
    if create_docker:
        build_env = f"{SCARGO_DOCKER_ENV}"
    else:
        build_env = "native"
    return build_env



# #
# @copyright Copyright (C) 2023 SpyroSoft Solutions S.A. All rights reserved.
# #
import os

COPYRIGHT = """# #
# Copyright (C) 2022 Spyrosoft Solutions. All rights reserved.
# #\n\n"""


def pop_and_add_subdirs(mock_folder):
    """pop_and_add_subdirs

    Args:
        mock_folder (_type_): _description_
    """
    mock_folder.pop()
    mock_folder_root = "/".join(map(str, mock_folder))

    try:
        listdir = os.listdir(mock_folder_root)
        subdirectories = [
            i for i in listdir if os.path.isdir(os.path.join(mock_folder_root, i))
        ]
        if "CMakeLists.txt" not in listdir:
            with open(
                f"{mock_folder_root}/CMakeLists.txt", "w", encoding="utf-8"
            ) as file:
                file.write(COPYRIGHT)
        with open(f"{mock_folder_root}/CMakeLists.txt", "r+", encoding="utf-8") as file:
            cmake_text = file.read()
            for i in subdirectories:
                if i not in cmake_text:
                    file.write(f"\nadd_subdirectory({i})")
    except OSError:
        pass


def add_subdirs_to_cmake(mock_root):
    """This function takes as input a path to mocks and a path to root.
    Then it creates a list of root directories where the index of "src"
    directory is found. The "start_path" is created which is a path of
    first "src_idx" directories. This path should be looped over to add
    subdirectories into CMakeLists.txt."""

    mock_folder = os.path.normpath(mock_root).split("/")

    if not os.path.exists(mock_root):
        mock_folder.pop()
    mock_folder_root = "/".join(map(str, mock_folder))

    try:
        listdir = os.listdir(mock_folder_root)
        subdirectories = [
            i for i in listdir if os.path.isdir(os.path.join(mock_folder_root, i))
        ]
        # check if CMake exists:
        # if CMakelists exists remove it from subdirectories list
        #  - this is a list only used to add subdirectories to the CMakeLists
        #  - this way it is also checked that CMakeLists exists in folder
        # if CMakelists doesn't exist create it in mock_folder
        if "CMakeLists.txt" not in listdir:
            with open(
                f"{mock_folder_root}/CMakeLists.txt", "w", encoding="utf-8"
            ) as file:
                file.write(COPYRIGHT)

        # check if all subidrectories exists in CMakeLists
        # if not add new ones
        with open(f"{mock_folder_root}/CMakeLists.txt", "r+", encoding="utf-8") as file:
            cmake_text = file.read()
            for i in subdirectories:
                if i not in cmake_text:
                    file.write(f"\nadd_subdirectory({i})")
        # if subdirectories is empty go folder up
        if not len(subdirectories) > 0:
            pop_and_add_subdirs(mock_folder)
        if not len(mock_folder) <= 2:
            pop_and_add_subdirs(mock_folder)
    except OSError:
        pass


def create_cmake_lists(mock_root, jinja_env):
    """This function copies a CMakeLists file from template folder to mock"""

    name = "mock_" + str(os.path.normpath(mock_root).split(os.path.sep)[-1])
    name = name.replace("_", " ").title().replace(" ", "")

    rendered = jinja_env.get_template("CMakeLists.txt").render(name=name)

    with open(f"{mock_root}/CMakeLists.txt", "w", encoding="utf-8") as file:
        file.write(rendered)



# #
# @copyright Copyright (C) 2023 SpyroSoft Solutions S.A. All rights reserved.
# #

"""Handle docker for project"""
import subprocess
import sys
from pathlib import Path

import docker

from scargo.scargo_src.sc_config import ProjectConfig
from scargo.scargo_src.sc_logger import get_logger
from scargo.scargo_src.sc_src import get_scargo_config_or_exit
from scargo.scargo_src.utils import get_project_root


def scargo_docker(
    build_docker: bool = False,
    no_cache: bool = False,
    run_docker: bool = False,
    exec_docker: bool = False,
):
    """
    :param bool run_docker: run command in docker
    :param bool exec_docker: connect to a running container
    :param bool build_docker: build docker
    :param bool no_cache: build docker without cache
    """
    project_path = get_project_root()

    # do not rebuild dockers in the docker
    if Path(project_path, ".dockerenv").exists():
        logger = get_logger()
        logger.error("Cannot used docker command inside the docker container.")
        sys.exit(1)

    config = get_scargo_config_or_exit()
    project_config = config.project

    docker_path = Path(project_path, ".devcontainer")

    if build_docker:
        _scargo_build_docker(docker_path, no_cache)

    if run_docker:
        _scargo_run_docker(docker_path, project_config)

    if exec_docker:
        _scargo_exec_docker(project_config)


def _scargo_build_docker(docker_path: Path, no_cache: bool = False) -> None:
    """
    Build docker

    :param Path docker_path: path to dockerfile
    :param bool no_cache: with/without docker cache
    :raises CalledProcessError: if docker build fail
    """
    logger = get_logger()
    logger.debug("Build docker environment.")

    cache = "--no-cache" if no_cache else ""
    try:
        subprocess.run(f"docker-compose build {cache}", shell=True, cwd=docker_path)
        logger.info("Initialize docker environment.")
    except subprocess.CalledProcessError:
        logger.error("Build docker fail.")


def _scargo_run_docker(docker_path: Path, project_config: ProjectConfig) -> None:
    """
    Run docker

    :param docker_path: path to docker
    :param project_config: project configuration
    :raises CalledProcessError: if docker did not start
    """
    logger = get_logger()
    logger.debug("Run docker environment.")

    try:
        subprocess.run(
            f"docker-compose run {project_config.name}_dev bash",
            shell=True,
            cwd=docker_path,
        )
        logger.info("Stop docker environment.")
    except subprocess.CalledProcessError:
        logger.error("Run docker fail.")


def _scargo_exec_docker(project_config: ProjectConfig):
    """
    Exec docker

    :param project_config: project configuration
    :raises CalledProcessError: if docker did not start
    """
    logger = get_logger()
    logger.debug("Exec docker environment.")

    image = project_config.docker_image_tag

    if not image:
        logger.error("docker-image-tag not defined in .toml under project section")
        sys.exit(1)

    client = docker.from_env()
    newest_container = client.containers.list(
        limit=1, filters={"ancestor": image, "status": "running"}
    )
    if not newest_container:
        logger.error("No running containers using image `%s` to attach to!", image)
        logger.info("Use scargo docker run to run container.")
        sys.exit(1)

    bash_command = ["bash"]
    cmd = ["docker", "exec", "-it", newest_container[0].id] + bash_command
    try:
        subprocess.run(cmd)
        logger.info("Stop exec docker environment.")
    except subprocess.CalledProcessError:
        logger.error("Exec docker fail.")
